import math
from copy import deepcopy

# TDOO: need to improve.
import numpy as np
from numpy import pi

from qutrunk.sim.local.pysim import Simulator
from qutrunk.tools.function_time import timefn
from .exceptions import LocalBackendError


# TODO: need to improve.
class MeasureResult:
    def __init__(self, id=0, value=0):
        # TODO: id是关键字，不建议使用
        self.id = id
        self.value = value


class OutcomeResult:
    def __init__(self, bit_str="", count=0):
        self.bitstr = bit_str
        self.count = count


class Result:
    def __init__(self):
        self.measureSet = []
        self.outcomeSet = []


class BackendLocalPython:
    def __init__(self):
        self.sim = Simulator()
        self.gate_map = {
            "H": "h",
            "P": "p",
            "CP": "cp",
            "R": "r",
            "Rx": "rx",
            "Rxx": "rxx",
            "Ry": "ry",
            "Ryy": "ryy",
            "Rz": "rz",
            "Rzz": "rzz",
            "NOT": "x",
            "X": "x",
            "Y": "y",
            "Z": "z",
            "S": "s",
            "T": "t",
            "Sdg": "sdg",
            "Tdg": "tdg",
            "SqrtSwap": "sqrtswap",
            "Swap": "swap",
            "CNOT": "cnot",
            "MCX": "cnot",
            "CY": "cy",
            "MCZ": "cz",
            "U3": "u3",
            "U2": "u2",
            "U1": "u1",
            "CRx": "crx",
            "CRy": "cry",
            "CRz": "crz",
            "X1": "x1",
            "Y1": "y1",
            "Z1": "z1",
            "CU1": "cu1",
            "CU3": "cu3",
            "U": "u",
            "CU": "cu",
            "CR": "cr",
            "iSwap": "iswap",
            "Barrier": "barrier",
            "SqrtX": "sprtx",
            "Id": "id",
            "CH": "ch",
            "SqrtXdg": "sqrtxdg",
            "CSqrtX": "csqrtx",
            "CSwap": "cswap",
            "AMP": "amp",
            "Matrix": "matrix"
        }
        self.cmds = []
        self.result = Result()

    @timefn
    def init(self, qubits):
        self.sim.create_qureg(qubits)
        self.sim.init_zero_state()

    @timefn
    def send_circuit(self, circuit, final):
        """
        Send the quantum circuit to local backend.

        Args:
            circuit: quantum circuit to send.
            final: True if quantum circuit finish, default False, \
                when final==True The backend program will release the computing resources.
        """
        start = circuit.cmd_cursor
        cmds = circuit.cmds[start:]
        for cmd in cmds:
            self.cmds.append(cmd)
            self.exec_cmd(cmd)
        if final:
            self.pack_result()

    @timefn
    def run(self, shots):
        """Run quantum circuit.

        Args:
            shots: circuit run times, for sampling.

        Returns:
            result: the Result object contain circuit running outcome.
        """
        # TODO: need to improve.
        run_times = shots - 1
        while run_times > 0:
            self.result.measureSet = []

            # reset state
            self.sim.init_zero_state()

            for cmd in self.cmds:
                self.exec_cmd(cmd)
            self.pack_result()
            run_times -= 1

        self.result.outcomeSet.sort(key=lambda a: a.bitstr)
        return self.result

    @timefn
    def get_prob(self, index):
        """
        Get the probability of a state-vector at an index in the full state vector.

        Args:
            index: index in state vector of probability amplitudes

        Returns:
            the probability of target index
        """
        return self.sim.get_prob(index)

    @timefn
    def get_probs(self, qubits):
        """Get all probabilities of circuit.

        Returns:
            An array contains all probabilities of circuit.
        """
        return self.sim.get_probs(qubits)

    @timefn
    def get_statevector(self):
        """Get state vector of circuit."""
        return self.sim.get_statevector()

    @timefn
    def qft(self, qubits):
        """
        Applies the quantum Fourier transform (QFT) to a specific subset of qubits of the register qureg

        Args:
            qubits: a list of the qubits to operate the QFT upon
        """
        if qubits:
            self.sim.apply_qft(qubits, len(qubits))
        else:
            self.sim.apply_full_qft()

    @timefn
    def get_expec_pauli_prod(self, pauli_prod_list):
        """
        Computes the expected value of a product of Pauli operators.

        Args:
            pauli_prod_list: a list contains the indices of the target qubits,\
                the Pauli codes (0=PAULI_I, 1=PAULI_X, 2=PAULI_Y, 3=PAULI_Z) to apply to the corresponding qubits.

        Returns:
            the expected value of a product of Pauli operators.
        """
        return self.sim.get_expec_pauli_prod(pauli_prod_list)

    @timefn
    def get_expec_pauli_sum(self, oper_type_list, term_coeff_list):
        """
        Computes the expected value of a sum of products of Pauli operators.

        Args:
            oper_type_list: a list of the Pauli codes (0=PAULI_I, 1=PAULI_X, 2=PAULI_Y, 3=PAULI_Z) \
                of all Paulis involved in the products of terms. A Pauli must be specified \
                for each qubit in the register, in every term of the sum.
            term_coeff_list: the coefficients of each term in the sum of Pauli products.

        Returns:
            the expected value of a sum of products of Pauli operators. 
        """
        return self.sim.get_expec_pauli_sum(oper_type_list, term_coeff_list)

    def exec_cmd(self, cmd):
        # TODO: need to improve.
        if str(cmd.gate) == "Measure":
            res = self.sim.measure(cmd.targets[0])
            mr = MeasureResult(cmd.targets[0], res)
            self.result.measureSet.append(mr)
            return

        return getattr(self, self.gate_map[str(cmd.gate)])(cmd)

    def pack_result(self):
        self.result.measureSet.sort(key=lambda a: a.id)
        bit_str = ""
        measureSet = self.result.measureSet
        for i in range(len(measureSet)):
            m = measureSet[i]
            bit_str += str(m.value)

        index = -1
        outcomeSet = self.result.outcomeSet
        for i in range(len(outcomeSet)):
            out = outcomeSet[i]
            if out.bitstr == bit_str:
                index = i
                break

        if index >= 0:
            self.result.outcomeSet[index].count += 1
        else:
            out = OutcomeResult(bit_str, 1)
            self.result.outcomeSet.append(out)

    def h(self, cmd):
        """the single-qubit Hadamard gate.

        Args:
           cmd: the Command object.
        """
        targets_len = len(cmd.targets)
        if targets_len != 1:
            raise LocalBackendError(
                f"h gate takes exactly one targets argument({targets_len} given)."
            )

        self.sim.hadamard(cmd.targets[0])

    def ch(self, cmd):
        """the controlled single-qubit Hadamard gate.

        Args:
           cmd: the Command object.

        # inverse is the same.
        """
        if len(cmd.targets) != 1 or len(cmd.controls) != 1:
            return

        factor = 1 / math.sqrt(2)
        # real part of complex number.
        ureal = np.array(
            [
                [1 * factor, 1 * factor],
                [1 * factor, -1 * factor],
            ]
        )

        # imaginary part of complex number.
        uimag = np.array(
            [
                [0, 0],
                [0, 0],
            ]
        )

        self.sim.ch(cmd.controls[0], cmd.targets[0], ureal, uimag)

    def cnot(self, cmd):
        if len(cmd.targets) == 1 and len(cmd.controls) == 1:
            self.sim.control_not(cmd.controls[0], cmd.targets[0])
        else:
            self.sim.multi_controlled_multi_qubit_not(
                cmd.controls, len(cmd.controls), cmd.targets, len(cmd.targets)
            )

    def p(self, cmd):
        """Shift the phase between |0> and |1> of a single qubit by a given angle.

        Args:
            cmd: the Command object.
        """
        t_num = len(cmd.targets)
        r_num = len(cmd.rotation)
        if (t_num != 1) or (r_num != 1):
            raise LocalBackendError(
                f"p gate takes exactly one targets and one rotation argument(targets:{t_num}, rotation:{r_num} given)."
            )

        rotation = cmd.rotation[0]
        if cmd.inverse:
            rotation = -rotation
        self.sim.phase_shift(cmd.targets[0], rotation)

    def cp(self, cmd):
        """Controlled-Phase gate.

        Args:
            cmd: the Command object.
        """
        t_num = len(cmd.targets)
        r_num = len(cmd.rotation)
        c_num = len(cmd.controls)
        if (t_num != 1) or (r_num != 1) or (c_num != 1):
            raise LocalBackendError(
                f"cp gate takes exactly one targets, one rotation and one controls argument(targets:{t_num}, "
                f"rotation:{r_num}, controls:{c_num} given)."
            )

        rotation = cmd.rotation[0]
        if cmd.inverse:
            rotation = -rotation
        self.sim.controlled_phase_shift(cmd.controls[0], cmd.targets[0], rotation)

    def r(self, cmd):
        """r gate.

        Args:
            cmd: the Command object.
        """
        t_num = len(cmd.targets)
        r_num = len(cmd.rotation)
        if (t_num != 1) or (r_num != 2):
            raise LocalBackendError(
                f"r gate takes exactly one targets and two rotation argument(targets:{t_num}, rotation:{r_num} given)."
            )

        theta = cmd.rotation[0]
        if cmd.inverse:
            theta = -theta
        phi = cmd.rotation[1]

        # real part of complex number
        ureal = np.array(
            [
                [math.cos(theta / 2), math.sin(-phi) * math.sin(theta / 2)],
                [math.sin(phi) * math.sin(theta / 2), math.cos(theta / 2)],
            ]
        )

        # imaginary part of complex number
        uimag = np.array(
            [
                [0, -1 * math.cos(-phi) * math.sin(theta / 2)],
                [-1 * math.cos(phi) * math.sin(theta / 2), 0],
            ]
        )

        self.sim.rotate(cmd.targets[0], ureal, uimag)

    def rx(self, cmd):
        """Rotate a single qubit by a given angle around the X-axis of the Bloch-sphere.

        Args:
            cmd: the Command object.
        """
        targets_len = len(cmd.targets)
        if targets_len != 1:
            raise LocalBackendError(
                f"rx gate takes exactly one targets argument({targets_len} given)."
            )

        rotation = cmd.rotation[0]
        if cmd.inverse:
            rotation = -rotation
        self.sim.rotate_x(cmd.targets[0], rotation)

    def rxx(self, cmd):
        """rxx gate.

        Args:
            cmd: the Command object.
        """
        targets_len = len(cmd.targets)
        if targets_len != 2:
            raise LocalBackendError(
                f"rxx gate takes exactly two targets argument({targets_len} given)."
            )

        angle = cmd.rotation[0]
        if cmd.inverse:
            angle = -angle

        ureal = np.array(
            [
                [math.cos(0.5 * angle), 0, 0, 0],
                [0, math.cos(0.5 * angle), 0, 0],
                [0, 0, math.cos(0.5 * angle), 0],
                [0, 0, 0, math.cos(0.5 * angle)],
            ]
        )
        uimag = np.array(
            [
                [0, 0, 0, -1 * math.sin(0.5 * angle)],
                [0, 0, -1 * math.sin(0.5 * angle), 0],
                [0, -1 * math.sin(0.5 * angle), 0, 0],
                [-1 * math.sin(0.5 * angle), 0, 0, 0],
            ]
        )

        self.sim.apply_matrix4(cmd.targets[0], cmd.targets[1], ureal, uimag)

    def ry(self, cmd):
        """ry gate.

        Args:
           cmd: the Command object.
        """
        targets_len = len(cmd.targets)
        if targets_len != 1:
            raise LocalBackendError(
                f"ry gate takes exactly one targets argument({targets_len} given)."
            )

        rotation = cmd.rotation[0]
        if cmd.inverse:
            rotation = -rotation

        self.sim.rotate_y(cmd.targets[0], rotation)

    def ryy(self, cmd):
        """ryy gate.

        Args:
            cmd: the Command object.
        """
        targets_len = len(cmd.targets)
        if targets_len != 2:
            raise LocalBackendError(
                f"ryy gate takes exactly two targets argument({targets_len} given)."
            )

        angle = cmd.rotation[0]
        if cmd.inverse:
            angle = -angle

        # real part of complex number
        ureal = np.array(
            [
                [math.cos(0.5 * angle), 0, 0, 0],
                [0, math.cos(0.5 * angle), 0, 0],
                [0, 0, math.cos(0.5 * angle), 0],
                [0, 0, 0, math.cos(0.5 * angle)],
            ]
        )
        # imaginary part of complex number
        uimag = np.array(
            [
                [0, 0, 0, 1 * math.sin(0.5 * angle)],
                [0, 0, -1 * math.sin(0.5 * angle), 0],
                [0, -1 * math.sin(0.5 * angle), 0, 0],
                [1 * math.sin(0.5 * angle), 0, 0, 0],
            ]
        )

        self.sim.apply_matrix4(cmd.targets[0], cmd.targets[1], ureal, uimag)

    def rz(self, cmd):
        """rz gate.

        Args:
            cmd: the Command object.
        """
        targets_len = len(cmd.targets)
        if targets_len != 1:
            raise LocalBackendError(
                f"rz gate takes exactly one targets argument({targets_len} given)."
            )

        rotation = cmd.rotation[0]
        if cmd.inverse:
            rotation = -rotation

        self.sim.rotate_z(cmd.targets[0], rotation)

    def rzz(self, cmd):
        """rzz gate.

        Args:
            cmd: the Command object.
        """
        targets_len = len(cmd.targets)
        if targets_len != 2:
            raise LocalBackendError(
                f"rzz gate takes exactly two targets argument({targets_len} given)."
            )

        angle = cmd.rotation[0]
        if cmd.inverse:
            angle = -angle

        # real part of complex number
        ureal = np.array(
            [
                [math.cos(0.5 * angle), 0, 0, 0],
                [0, math.cos(0.5 * angle), 0, 0],
                [0, 0, math.cos(0.5 * angle), 0],
                [0, 0, 0, math.cos(0.5 * angle)],
            ]
        )
        # imaginary part of complex number
        uimag = np.array(
            [
                [-1 * math.sin(0.5 * angle), 0, 0, 0],
                [0, math.sin(0.5 * angle), 0, 0],
                [0, 0, math.sin(0.5 * angle), 0],
                [0, 0, 0, -1 * math.sin(0.5 * angle)],
            ]
        )

        self.sim.apply_matrix4(cmd.targets[0], cmd.targets[1], ureal, uimag)

    def x(self, cmd):
        """The single-qubit Pauli-X gate.

        Args:
            cmd: the Command object.
        """
        targets_len = len(cmd.targets)
        if targets_len != 1:
            raise LocalBackendError(
                f"x gate takes exactly one targets argument({targets_len} given)."
            )

        self.sim.pauli_x(cmd.targets[0])

    def y(self, cmd):
        """
        The single-qubit Pauli-Y gate.

            Args:
                cmd: the Command object.
        """
        targets_len = len(cmd.targets)
        if targets_len != 1:
            raise LocalBackendError(
                f"y gate takes exactly one targets argument({targets_len} given)."
            )

        self.sim.pauli_y(cmd.targets[0])

    def z(self, cmd):
        """The single-qubit Pauli-Z gate.

        Args:
          cmd: the Command object.
        """
        targets_len = len(cmd.targets)
        if targets_len != 1:
            raise LocalBackendError(
                f"z gate takes exactly one targets argument({targets_len} given)."
            )

        self.sim.pauli_z(cmd.targets[0])

    def s(self, cmd):
        """the single-qubit S gate.

        Args:
          cmd: the Command object.
        """
        targets_len = len(cmd.targets)
        if targets_len != 1:
            raise LocalBackendError(
                f"s gate takes exactly one targets argument({targets_len} given)."
            )

        if cmd.inverse:
            cmd = deepcopy(cmd)
            cmd.inverse = False
            self.sdg(cmd)
        else:
            self.sim.s_gate(cmd.targets[0])

    def t(self, cmd):
        """the single-qubit T gate.

        Args:
          cmd: the Command object.
        """
        targets_len = len(cmd.targets)
        if targets_len != 1:
            raise LocalBackendError(
                f"t gate takes exactly one targets argument({targets_len} given)."
            )

        if cmd.inverse:
            cmd = deepcopy(cmd)
            cmd.inverse = False
            self.tdg(cmd)
        else:
            self.sim.t_gate(cmd.targets[0])

    def sdg(self, cmd):
        if cmd.inverse:
            cmd_temp = deepcopy(cmd)
            cmd_temp.inverse = False
            self.s(cmd_temp)
        else:
            ureal = np.array([[1, 0], [0, 0]])
            uimag = np.array([[0, 0], [0, -1]])
            self.sim.sdg(cmd.targets[0], ureal, uimag)

    def tdg(self, cmd):
        if cmd.inverse:
            cmd_temp = deepcopy(cmd)
            cmd_temp.inverse = False
            self.t(cmd_temp)
        else:
            ureal = np.array([[1, 0], [0, 1 / math.sqrt(2)]])
            uimag = np.array([[0, 0], [0, -1 / math.sqrt(2)]])
            self.sim.tdg(cmd.targets[0], ureal, uimag)

    def sqrtswap(self, cmd):
        targets_len = len(cmd.targets)
        if targets_len != 2:
            raise LocalBackendError(
                f"sqrtswap gate takes exactly one targets argument({targets_len} given)."
            )

        ureal = []
        uimag = []
        if cmd.inverse:
            ureal = np.array(
                [[1, 0, 0, 0], [0, 0.5, 0.5, 0], [0, 0.5, 0.5, 0], [0, 0, 0, 1]]
            )
            uimag = np.array(
                [[0, 0, 0, 0], [0, -0.5, 0.5, 0], [0, 0.5, -0.5, 0], [0, 0, 0, 0]]
            )
        else:
            ureal = np.array(
                [[1, 0, 0, 0], [0, 0.5, 0.5, 0], [0, 0.5, 0.5, 0], [0, 0, 0, 1]]
            )
            uimag = np.array(
                [[0, 0, 0, 0], [0, 0.5, -0.5, 0], [0, -0.5, 0.5, 0], [0, 0, 0, 0]]
            )
        self.sim.sqrtswap(cmd.targets[0], cmd.targets[1], ureal, uimag)

    def swap(self, cmd):
        # inverse is the same
        if len(cmd.targets) != 2:
            return

        self.sim.swap(cmd.targets[0], cmd.targets[1])

    def cy(self, cmd):
        # inverse is the same
        if len(cmd.targets) != 1 or len(cmd.controls) != 1:
            return

        self.sim.cy(cmd.controls[0], cmd.targets[0])

    def cz(self, cmd):
        # inverse is the same
        ctrls = cmd.controls[:]
        ctrls.extend(cmd.targets)

        self.sim.cz(ctrls, len(ctrls))

    def u3(self, cmd):
        if len(cmd.rotation) != 3 or len(cmd.targets) != 1:
            return

        theta = cmd.rotation[0]
        phi = cmd.rotation[1]
        lam = cmd.rotation[2]
        if cmd.inverse:
            theta1 = -theta
            phi1 = -lam
            lam1 = -phi
            theta = theta1
            phi = phi1
            lam = lam1

        ureal = np.array(
            [
                [math.cos(theta / 2), -1 * math.cos(lam) * math.sin(theta / 2)],
                [
                    math.cos(phi) * math.sin(theta / 2),
                    math.cos(phi + lam) * math.cos(theta / 2),
                ],
            ]
        )
        uimag = np.array(
            [
                [0, -1 * math.sin(lam) * math.sin(theta / 2)],
                [
                    math.sin(phi) * math.sin(theta / 2),
                    math.sin(phi + lam) * math.cos(theta / 2),
                ],
            ]
        )

        self.sim.u3(cmd.targets[0], ureal, uimag)

    def u2(self, cmd):
        if len(cmd.rotation) != 2 or len(cmd.targets) != 1:
            return

        phi = cmd.rotation[0]
        lam = cmd.rotation[1]

        if cmd.inverse:
            phi1 = -lam - pi
            lam1 = -phi + pi
            phi = phi1
            lam = lam1

        factor = 1 / math.sqrt(2)
        ureal = np.array(
            [
                [1 * factor, -factor * math.cos(lam)],
                [factor * math.cos(phi), factor * math.cos(phi + lam)],
            ]
        )
        uimag = np.array(
            [
                [0, -factor * math.sin(lam)],
                [factor * math.sin(phi), factor * math.sin(phi + lam)],
            ]
        )

        self.sim.u2(cmd.targets[0], ureal, uimag)

    def u1(self, cmd):
        if len(cmd.rotation) != 1 or len(cmd.targets) != 1:
            return

        alpha = cmd.rotation[0]
        if cmd.inverse:
            alpha = -alpha

        ureal = np.array([[1, 0], [0, math.cos(alpha)]])
        uimag = np.array([[0, 0], [0, math.sin(alpha)]])

        self.sim.u1(cmd.targets[0], ureal, uimag)

    def crx(self, cmd):
        if len(cmd.rotation) != 1 or len(cmd.targets) != 1 or len(cmd.controls) != 1:
            return

        rotation = cmd.rotation[0]
        if cmd.inverse:
            rotation = -rotation

        self.sim.crx(cmd.controls[0], cmd.targets[0], rotation)

    def cry(self, cmd):
        if len(cmd.rotation) != 1 or len(cmd.targets) != 1 or len(cmd.controls) != 1:
            return

        rotation = cmd.rotation[0]
        if cmd.inverse:
            rotation = -rotation

        self.sim.cry(cmd.controls[0], cmd.targets[0], rotation)

    def crz(self, cmd):
        if len(cmd.rotation) != 1 or len(cmd.targets) != 1 or len(cmd.controls) != 1:
            return

        rotation = cmd.rotation[0]
        if cmd.inverse:
            rotation = -rotation

        self.sim.crz(cmd.controls[0], cmd.targets[0], rotation)

    def x1(self, cmd):
        if len(cmd.targets) != 1:
            return

        ureal = []
        uimag = []
        if cmd.inverse:
            factor = math.sqrt(2)
            ureal = np.array([[0.5 * factor, 0], [0, 0.5 * factor]])
            uimag = np.array([[0, 0.5 * factor], [0.5 * factor, 0]])
        else:
            factor = 1 / math.sqrt(2)
            ureal = np.array([[1 * factor, 0], [0, 1 * factor]])
            uimag = np.array([[0, -1 * factor], [-1 * factor, 0]])

        self.sim.x1(cmd.targets[0], ureal, uimag)

    def y1(self, cmd):
        if len(cmd.targets) != 1:
            return

        ureal = []
        uimag = []
        if cmd.inverse:
            factor = math.sqrt(2)
            ureal = np.array(
                [[0.5 * factor, 0.5 * factor], [-0.5 * factor, 0.5 * factor]]
            )
            uimag = np.array([[0, 0], [0, 0]])
        else:
            factor = 1 / math.sqrt(2)
            ureal = np.array([[1 * factor, -1 * factor], [1 * factor, 1 * factor]])
            uimag = np.array([[0, 0], [0, 0]])

        self.sim.y1(cmd.targets[0], ureal, uimag)

    def z1(self, cmd):
        if len(cmd.targets) != 1:
            return

        rotation = pi / 4.0
        if cmd.inverse:
            rotation = -rotation

        ureal = np.array([[math.cos(-rotation), 0], [0, math.cos(rotation)]])
        uimag = np.array([[math.sin(-rotation), 0], [0, math.sin(rotation)]])

        self.sim.z1(cmd.targets[0], ureal, uimag)

    def cu1(self, cmd):
        if len(cmd.rotation) != 1 or len(cmd.targets) != 1 or len(cmd.controls) != 1:
            return

        alpha = cmd.rotation[0]

        if cmd.inverse:
            alpha = -alpha

        ureal = np.array(
            [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, math.cos(alpha)]]
        )
        uimag = np.array(
            [[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, math.sin(alpha)]]
        )

        self.sim.cu1(cmd.controls[0], cmd.targets[0], ureal, uimag)

    def cu3(self, cmd):
        if len(cmd.rotation) != 3 or len(cmd.targets) != 1 or len(cmd.controls) != 1:
            return

        theta = cmd.rotation[0]
        phi = cmd.rotation[1]
        lam = cmd.rotation[2]
        if cmd.inverse:
            theta1 = -theta
            phi1 = -lam
            lam1 = -phi
            theta = theta1
            phi = phi1
            lam = lam1

        ureal = np.array(
            [
                [1, 0, 0, 0],
                [0, math.cos(theta / 2), 0, -math.cos(lam) * math.sin(theta / 2)],
                [0, 0, 1, 0],
                [
                    0,
                    math.cos(phi) * math.sin(theta / 2),
                    0,
                    math.cos(phi + lam) * math.cos(theta / 2),
                ],
            ]
        )
        uimag = np.array(
            [
                [0, 0, 0, 0],
                [0, 0, 0, -math.sin(lam) * math.sin(theta / 2)],
                [0, 0, 0, 0],
                [
                    0,
                    math.sin(phi) * math.sin(theta / 2),
                    0,
                    math.sin(phi + lam) * math.cos(theta / 2),
                ],
            ]
        )

        self.sim.cu3(cmd.controls[0], cmd.targets[0], ureal, uimag)

    def u(self, cmd):
        self.u3(cmd)

    def cu(self, cmd):
        if len(cmd.rotation) != 4 or len(cmd.targets) != 1 or len(cmd.controls) != 1:
            return

        theta = cmd.rotation[0]
        phi = cmd.rotation[1]
        lam = cmd.rotation[2]
        gamma = cmd.rotation[3]
        if cmd.inverse:
            theta1 = -theta
            phi1 = -lam
            lam1 = -phi
            gamma1 = -gamma
            theta = theta1
            phi = phi1
            lam = lam1
            gamma = gamma1

        ureal = np.array(
            [
                [1, 0, 0, 0],
                [
                    0,
                    math.cos(gamma) * math.cos(theta / 2),
                    0,
                    -math.cos(gamma + lam) * math.sin(theta / 2),
                ],
                [0, 0, 1, 0],
                [
                    0,
                    math.cos(gamma + phi) * math.sin(theta / 2),
                    0,
                    math.cos(gamma + phi + lam) * math.cos(theta / 2),
                ],
            ]
        )
        uimag = np.array(
            [
                [0, 0, 0, 0],
                [
                    0,
                    math.sin(gamma) * math.cos(theta / 2),
                    0,
                    -math.sin(gamma + lam) * math.sin(theta / 2),
                ],
                [0, 0, 0, 0],
                [
                    0,
                    math.sin(gamma + phi) * math.sin(theta / 2),
                    0,
                    math.sin(gamma + phi + lam) * math.cos(theta / 2),
                ],
            ]
        )

        self.sim.cu(cmd.controls[0], cmd.targets[0], ureal, uimag)

    def cr(self, cmd):
        if len(cmd.rotation) != 1 or len(cmd.targets) != 1 or len(cmd.controls) != 1:
            return

        theta = cmd.rotation[0]
        if cmd.inverse:
            theta = -theta

        ureal = np.array(
            [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, math.cos(theta)]]
        )
        uimag = np.array(
            [[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, math.sin(theta)]]
        )

        self.sim.cr(cmd.controls[0], cmd.targets[0], ureal, uimag)

    def iswap(self, cmd):
        if len(cmd.targets) != 2:
            return

        ureal = []
        uimag = []
        if cmd.inverse:
            ureal = np.array(
                [
                    [1, 0, 0, 0],
                    [0, 0, 0, 0],
                    [0, 0, 0, 0],
                    [0, 0, 0, 1],
                ]
            )
            uimag = np.array(
                [
                    [0, 0, 0, 0],
                    [0, 0, -1, 0],
                    [0, -1, 0, 0],
                    [0, 0, 0, 0],
                ]
            )
        else:
            ureal = np.array(
                [
                    [1, 0, 0, 0],
                    [0, 0, 0, 0],
                    [0, 0, 0, 0],
                    [0, 0, 0, 1],
                ]
            )
            uimag = np.array(
                [
                    [0, 0, 0, 0],
                    [0, 0, 1, 0],
                    [0, 1, 0, 0],
                    [0, 0, 0, 0],
                ]
            )

        self.sim.iswap(cmd.targets[0], cmd.targets[1], ureal, uimag)

    def barrier(self, cmd):
        return

    def sprtx(self, cmd):
        if len(cmd.targets) != 1:
            return

        ureal = []
        uimag = []
        if cmd.inverse:
            ureal = np.array([[0.5, 0.5], [0.5, 0.5]])
            uimag = np.array([[-0.5, 0.5], [0.5, -0.5]])
        else:
            ureal = np.array([[0.5, 0.5], [0.5, 0.5]])
            uimag = np.array([[0.5, -0.5], [-0.5, 0.5]])

        self.sim.sqrtx(cmd.targets[0], ureal, uimag)

    def sqrtxdg(self, cmd):
        if len(cmd.targets) != 1:
            return

        ureal = []
        uimag = []
        if cmd.inverse:
            ureal = np.array([[0.5, 0.5], [0.5, 0.5]])
            uimag = np.array([[0.5, -0.5], [-0.5, 0.5]])
        else:
            ureal = np.array([[0.5, 0.5], [0.5, 0.5]])
            uimag = np.array([[-0.5, 0.5], [0.5, -0.5]])

        self.sim.sqrtxdg(cmd.targets[0], ureal, uimag)

    def csqrtx(self, cmd):
        if len(cmd.targets) != 1 or len(cmd.controls) != 1:
            return

        ureal = []
        uimag = []
        if cmd.inverse:
            ureal = np.array([[0.5, 0.5], [0.5, 0.5]])
            uimag = np.array([[-0.5, 0.5], [0.5, -0.5]])
        else:
            ureal = np.array([[0.5, 0.5], [0.5, 0.5]])
            uimag = np.array([[0.5, -0.5], [-0.5, 0.5]])

        self.sim.csqrtx(cmd.controls[0], cmd.targets[0], ureal, uimag)

    def cswap(self, cmd):
        if len(cmd.targets) != 2 or len(cmd.controls) != 1:
            return

        ureal = np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])
        uimag = np.array([[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]])

        self.sim.cswap(cmd.controls[0], cmd.targets[0], cmd.targets[1], ureal, uimag)

    def id(self, cmd):
        # do nothing
        pass

    def amp(self, cmd):
        """the set amplitudes gate.

        Args:
           cmd: the Command object.
        """

        self.sim.amp(cmd.cmdex.amp.reals, cmd.cmdex.amp.imags, cmd.cmdex.amp.startind, cmd.cmdex.amp.numamps)

    def matrix(self, cmd):
        """Apply custom matrix gate.

        Args:
           cmd: the Command object.
        """

        self.sim.matrix(cmd.controls, cmd.targets, cmd.cmdex.mat.reals, cmd.cmdex.mat.imags)



"""Quantum random number generator."""

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, Measure, All


def run_random_byte(backend=None):
    # Create circuit with local simulator
    qc = QCircuit(backend)

    # Allocate 8 quantuam qubits
    qureg = qc.allocate(8)

    # Apply quantum gates
    All(H) * qureg

    # Measure all quantum qubits
    All(Measure) * qureg

    # Print circuit
    qc.print()

    # Run circuit
    res = qc.run()

    # Print measure result like:
    # [0, 1, 0, 0, 0, 0, 1, 0]
    print(res.get_measure())

    return qc


if __name__ == "__main__":
    # Run locally
    circuit = run_random_byte()



"""Exception for errors raised while parsing OPENQASM."""
from qutrunk.exceptions import QuTrunkError


class QasmError(QuTrunkError):
    """Base class for errors raised while parsing OPENQASM."""

    def __init__(self, *message):
        """Set the error message."""
        super().__init__(*message)



import json
from collections import defaultdict
from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H,CNOT,Measure
qc = QCircuit()
qr = qc.allocate(2)
H | qr[0]
CNOT | (qr[0], qr[1])
Measure | qr[0]
Measure | qr[1]
outcome = qc.get_probs()
qc.run()
Tree = lambda: defaultdict(Tree)
tree = Tree()
for _, item in enumerate(outcome):
    i = str(bin(item['idx']))[2:]
    tree['probs'][i]['probability'] = item['prob']
    tree['probs'][i]['angle'] = 0
print(json.dumps(tree))
with open(file=r'result.json', mode='w', encoding='utf-8') as f:
    file = json.dump(json.dumps(tree), f)



#
# Autogenerated by Thrift Compiler (0.16.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
all_structs = []


class ErrCode(object):
    COM_SUCCESS = 0
    COM_OTHRE = 1
    COM_INVALID_PARAM = 2
    COM_TIMEOUT = 3
    COM_EXCEPTION = 4
    COM_NOT_EXIST = 5
    COM_IS_EXIST = 6
    COM_IS_QUEUE = 7
    COM_MEM_NOT_ENOUGH = 8
    COM_NOT_INIT = 9
    COM_NOT_REGISTER = 10
    COM_NOT_RESOURCE = 11
    COM_PRASE_ERROR = 12

    _VALUES_TO_NAMES = {
        0: "COM_SUCCESS",
        1: "COM_OTHRE",
        2: "COM_INVALID_PARAM",
        3: "COM_TIMEOUT",
        4: "COM_EXCEPTION",
        5: "COM_NOT_EXIST",
        6: "COM_IS_EXIST",
        7: "COM_IS_QUEUE",
        8: "COM_MEM_NOT_ENOUGH",
        9: "COM_NOT_INIT",
        10: "COM_NOT_REGISTER",
        11: "COM_NOT_RESOURCE",
        12: "COM_PRASE_ERROR",
    }

    _NAMES_TO_VALUES = {
        "COM_SUCCESS": 0,
        "COM_OTHRE": 1,
        "COM_INVALID_PARAM": 2,
        "COM_TIMEOUT": 3,
        "COM_EXCEPTION": 4,
        "COM_NOT_EXIST": 5,
        "COM_IS_EXIST": 6,
        "COM_IS_QUEUE": 7,
        "COM_MEM_NOT_ENOUGH": 8,
        "COM_NOT_INIT": 9,
        "COM_NOT_REGISTER": 10,
        "COM_NOT_RESOURCE": 11,
        "COM_PRASE_ERROR": 12,
    }


class BaseCode(object):
    """
    Attributes:
     - code
     - msg

    """


    def __init__(self, code=None, msg=None,):
        self.code = code
        self.msg = msg

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.I32:
                    self.code = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.msg = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('BaseCode')
        if self.code is not None:
            oprot.writeFieldBegin('code', TType.I32, 1)
            oprot.writeI32(self.code)
            oprot.writeFieldEnd()
        if self.msg is not None:
            oprot.writeFieldBegin('msg', TType.STRING, 2)
            oprot.writeString(self.msg.encode('utf-8') if sys.version_info[0] == 2 else self.msg)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        if self.code is None:
            raise TProtocolException(message='Required field code is unset!')
        if self.msg is None:
            raise TProtocolException(message='Required field msg is unset!')
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(BaseCode)
BaseCode.thrift_spec = (
    None,  # 0
    (1, TType.I32, 'code', None, None, ),  # 1
    (2, TType.STRING, 'msg', 'UTF8', None, ),  # 2
)
fix_spec(all_structs)
del all_structs



from enum import IntEnum

import pytest

from qutrunk.backends import BackendLocal, BackendQuSprout, BackendIBM, BackendAWSLocal, BackendAWSDevice
from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, CNOT, All, Measure


class BackendType(IntEnum):
    LOCAL = 1
    QU_SPROUT = 2
    QU_SAAS = 3
    IBM = 4
    AWS_LOCAL = 5
    AWS_DEVICE = 6


# all backend type
@pytest.fixture(params=[int(bt) for bt in list(BackendType)])
def backend_type(request):
    return request.param


@pytest.fixture
def circuit(backend_type):
    cir = QCircuit(backend=create_backend(backend_type))
    return cir


def create_backend(backend_type):
    if backend_type == BackendType.LOCAL:
        be = BackendLocal()
    elif backend_type == BackendType.QU_SPROUT:
        be = BackendQuSprout()
    elif backend_type == BackendType.IBM:
        be = BackendIBM()
    elif backend_type == BackendType.AWS_LOCAL:
        be = BackendAWSLocal()
    elif backend_type == BackendType.AWS_DEVICE:
        be = BackendAWSDevice()
    else:
        raise ValueError(f"unsupported backend type:{backend_type}")

    return be


# override default all backend_type
@pytest.mark.parametrize("backend_type", [BackendType.LOCAL, BackendType.AWS_LOCAL])
def test_backend(circuit):
    qr = circuit.allocate(2)
    H * qr[0]
    CNOT * (qr[0], qr[1])
    All(Measure) * qr
    print()
    circuit.draw()

    res = circuit.run(shots=100)
    print(circuit.backend.name)
    print(res.get_counts())



"""GHZ state example."""

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import CX, Measure, H, Barrier, All


def run_ghz(backend=None):
    # Create quantum circuit
    qc = QCircuit(name="ghz", backend=backend)

    # Allocate quantum qubits
    qr = qc.allocate(3)

    # Create a GHZ state
    H * qr[0]
    CX * (qr[0], qr[1])
    CX * (qr[0], qr[2])

    Barrier * qr

    # Measure all the qubits
    All(Measure) * qr

    # Run quantum circuit with 1024 times
    res = qc.run(shots=1024)

    # Print measure results like:
    # [{"000": 527}, {"111": 497}]
    print(res.get_counts())
    return qc


if __name__ == "__main__":
    # Run locally
    circuit = run_ghz()

    # Dram quantum circuit
    circuit.draw()




from .h import H, HGate, CH, CHGate
from .p import P, CP
from .r import R
from .rx import Rx, CRx
from .rxx import Rxx
from .ry import Ry, CRy
from .ryy import Ryy
from .rz import Rz, CRz
from .rzz import Rzz
from .swap import Swap, SwapGate, CSwap, CSwapGate
from .x import X, NOT, XGate, CX, CNOT, Toffoli, MCX, PauliX
from .y import Y, YGate, CY, CYGate, PauliY
from .z import Z, ZGate, CZ, MCZ, PauliZ
from .u1 import U1, CU1
from .u3 import U3, CU3, CU
from .u3 import U3 as U
from .barrier import Barrier, BarrierGate
from .basicgate import BasicGate, Observable, PauliType, PauliCoeff, PauliCoeffs
from .meta import All, Power, Gate
from .cr import CR
from .iswap import iSwap, iSwapGate
from .measure import Measure, MeasureGate
from .s import S, SGate
from .sdg import Sdg, SdgGate
from .sqrtswap import SqrtSwap, SqrtSwapGate
from .sqrtx import SqrtX, SqrtXGate, CSqrtX, CSqrtXGate
from .t import T, TGate
from .tdg import Tdg, TdgGate
from .u2 import U2
from .x1 import X1, X1Gate
from .y1 import Y1, Y1Gate
from .z1 import Z1, Z1Gate
from .i import I, IGate, PauliI
from .sxdg import SqrtXdg, SqrtXdgGate

CTRL_CNT_PARAM = "ctrl_cnt"



#
# Autogenerated by Thrift Compiler (0.16.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:package_prefix=qutrunk.sim.qusprout.
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys
import logging
from .ttypes import *
from thrift.Thrift import TProcessor
from thrift.transport import TTransport
all_structs = []


class Iface(object):
    def initQubits(self, req):
        """
        Parameters:
         - req

        """
        pass

    def sendCircuitCmd(self, req):
        """
        Parameters:
         - req

        """
        pass

    def cancelCmd(self, req):
        """
        Parameters:
         - req

        """
        pass

    def getProbAmp(self, req):
        """
        Parameters:
         - req

        """
        pass

    def getProbOfOutcome(self, req):
        """
        Parameters:
         - req

        """
        pass

    def getProbOfAllOutcome(self, req):
        """
        Parameters:
         - req

        """
        pass

    def getAllState(self, req):
        """
        Parameters:
         - req

        """
        pass

    def run(self, req):
        """
        Parameters:
         - req

        """
        pass

    def applyQFT(self, req):
        """
        Parameters:
         - req

        """
        pass

    def applyFullQFT(self, req):
        """
        Parameters:
         - req

        """
        pass

    def getExpecPauliProd(self, req):
        """
        Parameters:
         - req

        """
        pass

    def getExpecPauliSum(self, req):
        """
        Parameters:
         - req

        """
        pass


class Client(Iface):
    def __init__(self, iprot, oprot=None):
        self._iprot = self._oprot = iprot
        if oprot is not None:
            self._oprot = oprot
        self._seqid = 0

    def initQubits(self, req):
        """
        Parameters:
         - req

        """
        self.send_initQubits(req)
        return self.recv_initQubits()

    def send_initQubits(self, req):
        self._oprot.writeMessageBegin('initQubits', TMessageType.CALL, self._seqid)
        args = initQubits_args()
        args.req = req
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_initQubits(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = initQubits_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "initQubits failed: unknown result")

    def sendCircuitCmd(self, req):
        """
        Parameters:
         - req

        """
        self.send_sendCircuitCmd(req)
        return self.recv_sendCircuitCmd()

    def send_sendCircuitCmd(self, req):
        self._oprot.writeMessageBegin('sendCircuitCmd', TMessageType.CALL, self._seqid)
        args = sendCircuitCmd_args()
        args.req = req
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_sendCircuitCmd(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = sendCircuitCmd_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "sendCircuitCmd failed: unknown result")

    def cancelCmd(self, req):
        """
        Parameters:
         - req

        """
        self.send_cancelCmd(req)
        return self.recv_cancelCmd()

    def send_cancelCmd(self, req):
        self._oprot.writeMessageBegin('cancelCmd', TMessageType.CALL, self._seqid)
        args = cancelCmd_args()
        args.req = req
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_cancelCmd(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = cancelCmd_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "cancelCmd failed: unknown result")

    def getProbAmp(self, req):
        """
        Parameters:
         - req

        """
        self.send_getProbAmp(req)
        return self.recv_getProbAmp()

    def send_getProbAmp(self, req):
        self._oprot.writeMessageBegin('getProbAmp', TMessageType.CALL, self._seqid)
        args = getProbAmp_args()
        args.req = req
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_getProbAmp(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = getProbAmp_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "getProbAmp failed: unknown result")

    def getProbOfOutcome(self, req):
        """
        Parameters:
         - req

        """
        self.send_getProbOfOutcome(req)
        return self.recv_getProbOfOutcome()

    def send_getProbOfOutcome(self, req):
        self._oprot.writeMessageBegin('getProbOfOutcome', TMessageType.CALL, self._seqid)
        args = getProbOfOutcome_args()
        args.req = req
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_getProbOfOutcome(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = getProbOfOutcome_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "getProbOfOutcome failed: unknown result")

    def getProbOfAllOutcome(self, req):
        """
        Parameters:
         - req

        """
        self.send_getProbOfAllOutcome(req)
        return self.recv_getProbOfAllOutcome()

    def send_getProbOfAllOutcome(self, req):
        self._oprot.writeMessageBegin('getProbOfAllOutcome', TMessageType.CALL, self._seqid)
        args = getProbOfAllOutcome_args()
        args.req = req
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_getProbOfAllOutcome(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = getProbOfAllOutcome_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "getProbOfAllOutcome failed: unknown result")

    def getAllState(self, req):
        """
        Parameters:
         - req

        """
        self.send_getAllState(req)
        return self.recv_getAllState()

    def send_getAllState(self, req):
        self._oprot.writeMessageBegin('getAllState', TMessageType.CALL, self._seqid)
        args = getAllState_args()
        args.req = req
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_getAllState(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = getAllState_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "getAllState failed: unknown result")

    def run(self, req):
        """
        Parameters:
         - req

        """
        self.send_run(req)
        return self.recv_run()

    def send_run(self, req):
        self._oprot.writeMessageBegin('run', TMessageType.CALL, self._seqid)
        args = run_args()
        args.req = req
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_run(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = run_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "run failed: unknown result")

    def applyQFT(self, req):
        """
        Parameters:
         - req

        """
        self.send_applyQFT(req)
        return self.recv_applyQFT()

    def send_applyQFT(self, req):
        self._oprot.writeMessageBegin('applyQFT', TMessageType.CALL, self._seqid)
        args = applyQFT_args()
        args.req = req
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_applyQFT(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = applyQFT_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "applyQFT failed: unknown result")

    def applyFullQFT(self, req):
        """
        Parameters:
         - req

        """
        self.send_applyFullQFT(req)
        return self.recv_applyFullQFT()

    def send_applyFullQFT(self, req):
        self._oprot.writeMessageBegin('applyFullQFT', TMessageType.CALL, self._seqid)
        args = applyFullQFT_args()
        args.req = req
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_applyFullQFT(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = applyFullQFT_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "applyFullQFT failed: unknown result")

    def getExpecPauliProd(self, req):
        """
        Parameters:
         - req

        """
        self.send_getExpecPauliProd(req)
        return self.recv_getExpecPauliProd()

    def send_getExpecPauliProd(self, req):
        self._oprot.writeMessageBegin('getExpecPauliProd', TMessageType.CALL, self._seqid)
        args = getExpecPauliProd_args()
        args.req = req
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_getExpecPauliProd(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = getExpecPauliProd_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "getExpecPauliProd failed: unknown result")

    def getExpecPauliSum(self, req):
        """
        Parameters:
         - req

        """
        self.send_getExpecPauliSum(req)
        return self.recv_getExpecPauliSum()

    def send_getExpecPauliSum(self, req):
        self._oprot.writeMessageBegin('getExpecPauliSum', TMessageType.CALL, self._seqid)
        args = getExpecPauliSum_args()
        args.req = req
        args.write(self._oprot)
        self._oprot.writeMessageEnd()
        self._oprot.trans.flush()

    def recv_getExpecPauliSum(self):
        iprot = self._iprot
        (fname, mtype, rseqid) = iprot.readMessageBegin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(iprot)
            iprot.readMessageEnd()
            raise x
        result = getExpecPauliSum_result()
        result.read(iprot)
        iprot.readMessageEnd()
        if result.success is not None:
            return result.success
        raise TApplicationException(TApplicationException.MISSING_RESULT, "getExpecPauliSum failed: unknown result")


class Processor(Iface, TProcessor):
    def __init__(self, handler):
        self._handler = handler
        self._processMap = {}
        self._processMap["initQubits"] = Processor.process_initQubits
        self._processMap["sendCircuitCmd"] = Processor.process_sendCircuitCmd
        self._processMap["cancelCmd"] = Processor.process_cancelCmd
        self._processMap["getProbAmp"] = Processor.process_getProbAmp
        self._processMap["getProbOfOutcome"] = Processor.process_getProbOfOutcome
        self._processMap["getProbOfAllOutcome"] = Processor.process_getProbOfAllOutcome
        self._processMap["getAllState"] = Processor.process_getAllState
        self._processMap["run"] = Processor.process_run
        self._processMap["applyQFT"] = Processor.process_applyQFT
        self._processMap["applyFullQFT"] = Processor.process_applyFullQFT
        self._processMap["getExpecPauliProd"] = Processor.process_getExpecPauliProd
        self._processMap["getExpecPauliSum"] = Processor.process_getExpecPauliSum
        self._on_message_begin = None

    def on_message_begin(self, func):
        self._on_message_begin = func

    def process(self, iprot, oprot):
        (name, type, seqid) = iprot.readMessageBegin()
        if self._on_message_begin:
            self._on_message_begin(name, type, seqid)
        if name not in self._processMap:
            iprot.skip(TType.STRUCT)
            iprot.readMessageEnd()
            x = TApplicationException(TApplicationException.UNKNOWN_METHOD, 'Unknown function %s' % (name))
            oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
            x.write(oprot)
            oprot.writeMessageEnd()
            oprot.trans.flush()
            return
        else:
            self._processMap[name](self, seqid, iprot, oprot)
        return True

    def process_initQubits(self, seqid, iprot, oprot):
        args = initQubits_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = initQubits_result()
        try:
            result.success = self._handler.initQubits(args.req)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("initQubits", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_sendCircuitCmd(self, seqid, iprot, oprot):
        args = sendCircuitCmd_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = sendCircuitCmd_result()
        try:
            result.success = self._handler.sendCircuitCmd(args.req)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("sendCircuitCmd", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_cancelCmd(self, seqid, iprot, oprot):
        args = cancelCmd_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = cancelCmd_result()
        try:
            result.success = self._handler.cancelCmd(args.req)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("cancelCmd", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_getProbAmp(self, seqid, iprot, oprot):
        args = getProbAmp_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = getProbAmp_result()
        try:
            result.success = self._handler.getProbAmp(args.req)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("getProbAmp", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_getProbOfOutcome(self, seqid, iprot, oprot):
        args = getProbOfOutcome_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = getProbOfOutcome_result()
        try:
            result.success = self._handler.getProbOfOutcome(args.req)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("getProbOfOutcome", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_getProbOfAllOutcome(self, seqid, iprot, oprot):
        args = getProbOfAllOutcome_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = getProbOfAllOutcome_result()
        try:
            result.success = self._handler.getProbOfAllOutcome(args.req)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("getProbOfAllOutcome", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_getAllState(self, seqid, iprot, oprot):
        args = getAllState_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = getAllState_result()
        try:
            result.success = self._handler.getAllState(args.req)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("getAllState", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_run(self, seqid, iprot, oprot):
        args = run_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = run_result()
        try:
            result.success = self._handler.run(args.req)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("run", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_applyQFT(self, seqid, iprot, oprot):
        args = applyQFT_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = applyQFT_result()
        try:
            result.success = self._handler.applyQFT(args.req)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("applyQFT", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_applyFullQFT(self, seqid, iprot, oprot):
        args = applyFullQFT_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = applyFullQFT_result()
        try:
            result.success = self._handler.applyFullQFT(args.req)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("applyFullQFT", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_getExpecPauliProd(self, seqid, iprot, oprot):
        args = getExpecPauliProd_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = getExpecPauliProd_result()
        try:
            result.success = self._handler.getExpecPauliProd(args.req)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("getExpecPauliProd", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

    def process_getExpecPauliSum(self, seqid, iprot, oprot):
        args = getExpecPauliSum_args()
        args.read(iprot)
        iprot.readMessageEnd()
        result = getExpecPauliSum_result()
        try:
            result.success = self._handler.getExpecPauliSum(args.req)
            msg_type = TMessageType.REPLY
        except TTransport.TTransportException:
            raise
        except TApplicationException as ex:
            logging.exception('TApplication exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = ex
        except Exception:
            logging.exception('Unexpected exception in handler')
            msg_type = TMessageType.EXCEPTION
            result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
        oprot.writeMessageBegin("getExpecPauliSum", msg_type, seqid)
        result.write(oprot)
        oprot.writeMessageEnd()
        oprot.trans.flush()

# HELPER FUNCTIONS AND STRUCTURES


class initQubits_args(object):
    """
    Attributes:
     - req

    """


    def __init__(self, req=None,):
        self.req = req

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.req = qutrunk.sim.qusprout.qusproutdata.ttypes.InitQubitsReq()
                    self.req.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('initQubits_args')
        if self.req is not None:
            oprot.writeFieldBegin('req', TType.STRUCT, 1)
            self.req.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(initQubits_args)
initQubits_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'req', [qutrunk.sim.qusprout.qusproutdata.ttypes.InitQubitsReq, None], None, ),  # 1
)


class initQubits_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRUCT:
                    self.success = qutrunk.sim.qusprout.qusproutdata.ttypes.InitQubitsResp()
                    self.success.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('initQubits_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRUCT, 0)
            self.success.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(initQubits_result)
initQubits_result.thrift_spec = (
    (0, TType.STRUCT, 'success', [qutrunk.sim.qusprout.qusproutdata.ttypes.InitQubitsResp, None], None, ),  # 0
)


class sendCircuitCmd_args(object):
    """
    Attributes:
     - req

    """


    def __init__(self, req=None,):
        self.req = req

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.req = qutrunk.sim.qusprout.qusproutdata.ttypes.SendCircuitCmdReq()
                    self.req.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('sendCircuitCmd_args')
        if self.req is not None:
            oprot.writeFieldBegin('req', TType.STRUCT, 1)
            self.req.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(sendCircuitCmd_args)
sendCircuitCmd_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'req', [qutrunk.sim.qusprout.qusproutdata.ttypes.SendCircuitCmdReq, None], None, ),  # 1
)


class sendCircuitCmd_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRUCT:
                    self.success = qutrunk.sim.qusprout.qusproutdata.ttypes.SendCircuitCmdResp()
                    self.success.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('sendCircuitCmd_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRUCT, 0)
            self.success.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(sendCircuitCmd_result)
sendCircuitCmd_result.thrift_spec = (
    (0, TType.STRUCT, 'success', [qutrunk.sim.qusprout.qusproutdata.ttypes.SendCircuitCmdResp, None], None, ),  # 0
)


class cancelCmd_args(object):
    """
    Attributes:
     - req

    """


    def __init__(self, req=None,):
        self.req = req

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.req = qutrunk.sim.qusprout.qusproutdata.ttypes.CancelCmdReq()
                    self.req.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('cancelCmd_args')
        if self.req is not None:
            oprot.writeFieldBegin('req', TType.STRUCT, 1)
            self.req.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(cancelCmd_args)
cancelCmd_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'req', [qutrunk.sim.qusprout.qusproutdata.ttypes.CancelCmdReq, None], None, ),  # 1
)


class cancelCmd_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRUCT:
                    self.success = qutrunk.sim.qusprout.qusproutdata.ttypes.CancelCmdResp()
                    self.success.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('cancelCmd_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRUCT, 0)
            self.success.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(cancelCmd_result)
cancelCmd_result.thrift_spec = (
    (0, TType.STRUCT, 'success', [qutrunk.sim.qusprout.qusproutdata.ttypes.CancelCmdResp, None], None, ),  # 0
)


class getProbAmp_args(object):
    """
    Attributes:
     - req

    """


    def __init__(self, req=None,):
        self.req = req

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.req = qutrunk.sim.qusprout.qusproutdata.ttypes.GetProbAmpReq()
                    self.req.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getProbAmp_args')
        if self.req is not None:
            oprot.writeFieldBegin('req', TType.STRUCT, 1)
            self.req.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getProbAmp_args)
getProbAmp_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'req', [qutrunk.sim.qusprout.qusproutdata.ttypes.GetProbAmpReq, None], None, ),  # 1
)


class getProbAmp_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRUCT:
                    self.success = qutrunk.sim.qusprout.qusproutdata.ttypes.GetProbAmpResp()
                    self.success.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getProbAmp_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRUCT, 0)
            self.success.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getProbAmp_result)
getProbAmp_result.thrift_spec = (
    (0, TType.STRUCT, 'success', [qutrunk.sim.qusprout.qusproutdata.ttypes.GetProbAmpResp, None], None, ),  # 0
)


class getProbOfOutcome_args(object):
    """
    Attributes:
     - req

    """


    def __init__(self, req=None,):
        self.req = req

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.req = qutrunk.sim.qusprout.qusproutdata.ttypes.GetProbOfOutcomeReq()
                    self.req.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getProbOfOutcome_args')
        if self.req is not None:
            oprot.writeFieldBegin('req', TType.STRUCT, 1)
            self.req.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getProbOfOutcome_args)
getProbOfOutcome_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'req', [qutrunk.sim.qusprout.qusproutdata.ttypes.GetProbOfOutcomeReq, None], None, ),  # 1
)


class getProbOfOutcome_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRUCT:
                    self.success = qutrunk.sim.qusprout.qusproutdata.ttypes.GetProbOfOutcomeResp()
                    self.success.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getProbOfOutcome_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRUCT, 0)
            self.success.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getProbOfOutcome_result)
getProbOfOutcome_result.thrift_spec = (
    (0, TType.STRUCT, 'success', [qutrunk.sim.qusprout.qusproutdata.ttypes.GetProbOfOutcomeResp, None], None, ),  # 0
)


class getProbOfAllOutcome_args(object):
    """
    Attributes:
     - req

    """


    def __init__(self, req=None,):
        self.req = req

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.req = qutrunk.sim.qusprout.qusproutdata.ttypes.GetProbOfAllOutcomReq()
                    self.req.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getProbOfAllOutcome_args')
        if self.req is not None:
            oprot.writeFieldBegin('req', TType.STRUCT, 1)
            self.req.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getProbOfAllOutcome_args)
getProbOfAllOutcome_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'req', [qutrunk.sim.qusprout.qusproutdata.ttypes.GetProbOfAllOutcomReq, None], None, ),  # 1
)


class getProbOfAllOutcome_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRUCT:
                    self.success = qutrunk.sim.qusprout.qusproutdata.ttypes.GetProbOfAllOutcomResp()
                    self.success.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getProbOfAllOutcome_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRUCT, 0)
            self.success.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getProbOfAllOutcome_result)
getProbOfAllOutcome_result.thrift_spec = (
    (0, TType.STRUCT, 'success', [qutrunk.sim.qusprout.qusproutdata.ttypes.GetProbOfAllOutcomResp, None], None, ),  # 0
)


class getAllState_args(object):
    """
    Attributes:
     - req

    """


    def __init__(self, req=None,):
        self.req = req

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.req = qutrunk.sim.qusprout.qusproutdata.ttypes.GetAllStateReq()
                    self.req.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getAllState_args')
        if self.req is not None:
            oprot.writeFieldBegin('req', TType.STRUCT, 1)
            self.req.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getAllState_args)
getAllState_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'req', [qutrunk.sim.qusprout.qusproutdata.ttypes.GetAllStateReq, None], None, ),  # 1
)


class getAllState_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRUCT:
                    self.success = qutrunk.sim.qusprout.qusproutdata.ttypes.GetAllStateResp()
                    self.success.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getAllState_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRUCT, 0)
            self.success.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getAllState_result)
getAllState_result.thrift_spec = (
    (0, TType.STRUCT, 'success', [qutrunk.sim.qusprout.qusproutdata.ttypes.GetAllStateResp, None], None, ),  # 0
)


class run_args(object):
    """
    Attributes:
     - req

    """


    def __init__(self, req=None,):
        self.req = req

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.req = qutrunk.sim.qusprout.qusproutdata.ttypes.RunCircuitReq()
                    self.req.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('run_args')
        if self.req is not None:
            oprot.writeFieldBegin('req', TType.STRUCT, 1)
            self.req.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(run_args)
run_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'req', [qutrunk.sim.qusprout.qusproutdata.ttypes.RunCircuitReq, None], None, ),  # 1
)


class run_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRUCT:
                    self.success = qutrunk.sim.qusprout.qusproutdata.ttypes.RunCircuitResp()
                    self.success.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('run_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRUCT, 0)
            self.success.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(run_result)
run_result.thrift_spec = (
    (0, TType.STRUCT, 'success', [qutrunk.sim.qusprout.qusproutdata.ttypes.RunCircuitResp, None], None, ),  # 0
)


class applyQFT_args(object):
    """
    Attributes:
     - req

    """


    def __init__(self, req=None,):
        self.req = req

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.req = qutrunk.sim.qusprout.qusproutdata.ttypes.ApplyQFTReq()
                    self.req.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('applyQFT_args')
        if self.req is not None:
            oprot.writeFieldBegin('req', TType.STRUCT, 1)
            self.req.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(applyQFT_args)
applyQFT_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'req', [qutrunk.sim.qusprout.qusproutdata.ttypes.ApplyQFTReq, None], None, ),  # 1
)


class applyQFT_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRUCT:
                    self.success = qutrunk.sim.qusprout.qusproutdata.ttypes.ApplyQFTResp()
                    self.success.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('applyQFT_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRUCT, 0)
            self.success.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(applyQFT_result)
applyQFT_result.thrift_spec = (
    (0, TType.STRUCT, 'success', [qutrunk.sim.qusprout.qusproutdata.ttypes.ApplyQFTResp, None], None, ),  # 0
)


class applyFullQFT_args(object):
    """
    Attributes:
     - req

    """


    def __init__(self, req=None,):
        self.req = req

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.req = qutrunk.sim.qusprout.qusproutdata.ttypes.ApplyFullQFTReq()
                    self.req.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('applyFullQFT_args')
        if self.req is not None:
            oprot.writeFieldBegin('req', TType.STRUCT, 1)
            self.req.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(applyFullQFT_args)
applyFullQFT_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'req', [qutrunk.sim.qusprout.qusproutdata.ttypes.ApplyFullQFTReq, None], None, ),  # 1
)


class applyFullQFT_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRUCT:
                    self.success = qutrunk.sim.qusprout.qusproutdata.ttypes.ApplyFullQFTResp()
                    self.success.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('applyFullQFT_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRUCT, 0)
            self.success.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(applyFullQFT_result)
applyFullQFT_result.thrift_spec = (
    (0, TType.STRUCT, 'success', [qutrunk.sim.qusprout.qusproutdata.ttypes.ApplyFullQFTResp, None], None, ),  # 0
)


class getExpecPauliProd_args(object):
    """
    Attributes:
     - req

    """


    def __init__(self, req=None,):
        self.req = req

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.req = qutrunk.sim.qusprout.qusproutdata.ttypes.GetExpecPauliProdReq()
                    self.req.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getExpecPauliProd_args')
        if self.req is not None:
            oprot.writeFieldBegin('req', TType.STRUCT, 1)
            self.req.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getExpecPauliProd_args)
getExpecPauliProd_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'req', [qutrunk.sim.qusprout.qusproutdata.ttypes.GetExpecPauliProdReq, None], None, ),  # 1
)


class getExpecPauliProd_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRUCT:
                    self.success = qutrunk.sim.qusprout.qusproutdata.ttypes.GetExpecPauliProdResp()
                    self.success.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getExpecPauliProd_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRUCT, 0)
            self.success.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getExpecPauliProd_result)
getExpecPauliProd_result.thrift_spec = (
    (0, TType.STRUCT, 'success', [qutrunk.sim.qusprout.qusproutdata.ttypes.GetExpecPauliProdResp, None], None, ),  # 0
)


class getExpecPauliSum_args(object):
    """
    Attributes:
     - req

    """


    def __init__(self, req=None,):
        self.req = req

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.req = qutrunk.sim.qusprout.qusproutdata.ttypes.GetExpecPauliSumReq()
                    self.req.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getExpecPauliSum_args')
        if self.req is not None:
            oprot.writeFieldBegin('req', TType.STRUCT, 1)
            self.req.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getExpecPauliSum_args)
getExpecPauliSum_args.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'req', [qutrunk.sim.qusprout.qusproutdata.ttypes.GetExpecPauliSumReq, None], None, ),  # 1
)


class getExpecPauliSum_result(object):
    """
    Attributes:
     - success

    """


    def __init__(self, success=None,):
        self.success = success

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 0:
                if ftype == TType.STRUCT:
                    self.success = qutrunk.sim.qusprout.qusproutdata.ttypes.GetExpecPauliSumResp()
                    self.success.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('getExpecPauliSum_result')
        if self.success is not None:
            oprot.writeFieldBegin('success', TType.STRUCT, 0)
            self.success.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(getExpecPauliSum_result)
getExpecPauliSum_result.thrift_spec = (
    (0, TType.STRUCT, 'success', [qutrunk.sim.qusprout.qusproutdata.ttypes.GetExpecPauliSumResp, None], None, ),  # 0
)
fix_spec(all_structs)
del all_structs



from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, Measure, CNOT


def test_qusl_file():
    qc = QCircuit()
    qreg = qc.allocate(2)

    H * qreg[0]
    CNOT * (qreg[0], qreg[1])
    Measure * qreg[0]
    Measure * qreg[1]

    qc.print("a.qusl")

    expect_in = qc.cmds
    circuit_in = qc.load("a.qusl", "qusl")
    # circuit_in = qc.from_qusl_file(file="a.qusl")

    assert expect_in == circuit_in.cmds





# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""OPENQASM circuit object."""
from .exceptions import QasmError
from .qasmparser import QasmParser


class Qasm:
    """OPENQASM circuit object."""

    def __init__(self, filename=None, data=None):
        """Create an OPENQASM circuit object."""
        if filename is None and data is None:
            raise QasmError("Missing input file and/or data")
        if filename is not None and data is not None:
            raise QasmError(
                "File and data must not both be specified initializing qasm"
            )
        self._filename = filename
        self._data = data

    def return_filename(self):
        """Return the filename."""
        return self._filename

    def generate_tokens(self):
        """Returns a generator of the tokens."""
        if self._filename:
            with open(self._filename) as ifile:
                self._data = ifile.read()

        with QasmParser(self._filename) as qasm_p:
            return qasm_p.read_tokens()

    def parse(self):
        """Parse the data."""
        if self._filename:
            with open(self._filename) as ifile:
                self._data = ifile.read()

        with QasmParser(self._filename) as qasm_p:
            qasm_p.parse_debug(False)
            return qasm_p.parse(self._data)



import pytest

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H
from qutrunk.backends import BackendQuSprout
from check_all_state import check_all_state
from check_all_state_inverse import check_all_state_inverse


def test_h_gate():
    """测试H门"""
    # 使用本地量子计算模拟器
    circuit = QCircuit()
    qr = circuit.allocate(1)
    H * qr[0]
    res = circuit.get_statevector()

    # 使用BackendQuSprout量子计算模拟器
    circuit_box = QCircuit(backend=BackendQuSprout())
    qr_box = circuit_box.allocate(1)
    H * (qr_box[0])
    res_box = circuit_box.get_statevector()

    # 检查数据是否一致
    assert check_all_state(res, res_box)


def test_h_inverse_gate():
    """测试反转电路"""
    # 使用本地量子计算模拟器
    circuit = QCircuit()
    qr = circuit.allocate(1)
    # 获取原始数据
    org_res = circuit.get_statevector()

    # 进行逆操作
    H * qr[0]
    H * qr[0]
    circuit.cmds[1].inverse = True

    # 获取逆操作后数据
    final_res = circuit.get_statevector()

    # 检查逆操作前后数据是否一致
    assert check_all_state_inverse(org_res, final_res)


if __name__ == "__main__":
    """运行test文件"""
    pytest.main(["-v", "-s", "./test_h_gate.py"])



from abc import ABC, abstractmethod


class Backend(ABC):
    """Basic simulator: All simulators are derived from this class."""

    @abstractmethod
    def send_circuit(self, circuit, final=False):
        """Send the quantum circuit to backend.

        Args:
            circuit: Quantum circuit to send.
            final: True if quantum circuit finish, default False, \
                when final==True The backend program will release the computing resources.
        """
        raise NotImplementedError

    @abstractmethod
    def name(self):
        raise NotImplementedError



"""
QASM转化为qutrunk量子程序  convert_qasm_to_qutrunk
主要步骤：
1 读取包含QASM内容的文件
2 调用convert_qasm_to_qutrunk()方法
3 调用convert_qutrunk_to_qasm()方法以把量子程序转为QASM，通过比较输入和生成的QASM是否相同，判断QASM是否正确转化成量子程序。
"""

from .qasm import Qasm



"""Quantum Fourier Transform examples."""

from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import All, Measure
from qutrunk.circuit.ops import QFT


def full_qft():
    circuit = QCircuit()
    qreg = circuit.allocate(5)

    QFT * qreg

    circuit.draw(line_length=1000)
    state = circuit.get_statevector()
    print(state)

    All(Measure) * qreg

    res = circuit.run(shots=1000)
    print(res.get_counts())


if __name__ == "__main__":
    full_qft()



import pytest

from qutrunk.backends import BackendQuSprout
from qutrunk.circuit import QCircuit
from qutrunk.circuit.gates import H, Ry, PauliZ, PauliI, PauliX, PauliCoeff, PauliType, PauliCoeffs
from qutrunk.test.tools.float import equal

def test_expval_pauli_sum():
    circuit = QCircuit()
    q = circuit.allocate(2)

    H * q[0]
    Ry(1.23) * q[1]
    pauli_coeffs = PauliCoeffs() << PauliCoeff(0.12, [PauliType.PAULI_Z]) \
        << PauliCoeff(0.34, [PauliType.PAULI_X, PauliType.PAULI_I])
    expect_sum = circuit.expval_pauli_sum(pauli_coeffs)

    # circuit2 = QCircuit(backend=BackendQuSprout())
    # q2 = circuit2.allocate(2)
    # circuit2.set_cmds(circuit.cmds)
    # expect_sum2 = circuit2.expval_pauli_sum(pauli_coeffs)

    # 确保前后的计算结果是一致的, 对于上述例子后端计算的结果为0.34
    assert equal(expect_sum, 0.34)



"""
The entry for drawing quantum circuit diagram.
"""
from .utils import _get_instructions
from .text import TextDrawing


def circuit_drawer(circuit, output, line_length):
    """Draw the quantum circuit.

    According to the parameters, draw the quantum circuit.

    Args:
        circuit: The quantum circuit to draw.
        output: Select the output method to use for drawing the circuit.
        line_length: The length of the line.

    Returns:
        Drawn string.
    """
    if output == "text":
        return __text_circuit_drawer(circuit=circuit, line_length=line_length)


def __text_circuit_drawer(
    circuit, plot_barriers=True, vertical_compression="high", line_length=None
):
    """Draw quantum circuit by ascii art.

    Args:
        circuit: Input circuit.
        plot_barriers: Whether to export barriers.
        vertical_compression: Vertical compression with text (default: Medium).
        line_length: The maximum length of a line when drawing.

    Returns:
        Return the draw result.

    """
    qubits, cbits, nodes = _get_instructions(circuit)
    text_drawing = TextDrawing(qubits=qubits, cbits=cbits, nodes=nodes, circuit=circuit)
    text_drawing.plot_barriers = plot_barriers
    text_drawing.line_length = line_length
    text_drawing.vertical_compression = vertical_compression

    return text_drawing



import numpy as np

from qutrunk.circuit import Command
from .basicgate import BasicGate, Observable, PauliType
from qutrunk.circuit.qubit import QuBit


class YGate(BasicGate, Observable):
    """Apply the single-qubit Pauli-Y (also known as the Y or sigma-Y) gate.

    Example:
        .. code-block:: python

            Y * qr[0]
    """

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "Y"

    def __or__(self, qubit):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to apply Y gate.

        Example:
            .. code-block:: python

                Y * qr[0]

        Raises:
            NotImplementedError: If the argument is not a Qubit object.
        """
        if not isinstance(qubit, QuBit):
            raise TypeError("The argument must be Qubit object.")

        targets = [qubit.index]
        cmd = Command(self, targets, inverse=self.is_inverse)
        self.commit(qubit.circuit, cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, \
            reuse __or__ operator implement."""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.array([[0, -1j], [1j, 0]])

    def __call__(self, target):
        """
        Get Observable data.

        Args:
            target: The observed qubit.

        Returns:
            The observed data list, each item contains op type and target qubit, \
                e.g: [{"oper_type": 1, "target": 0}].
        """
        pauli = {}
        pauli["oper_type"] = PauliType.PAULI_Y.value
        pauli["target"] = target.index
        return pauli

    def inv(self):
        """Apply inverse gate."""
        gate = YGate()
        gate.is_inverse = not self.is_inverse
        return gate

    def ctrl(self, ctrl_cnt=1):
        """Apply controlled gate.
        
        Args:
            ctrl_cnt: The number of control qubits, default: 1.
        """
        if ctrl_cnt > 1:
            raise ValueError("Y gate do not support multiple control bits.")
        gate = CYGate()
        gate.is_inverse = self.is_inverse
        return gate


PauliY = Y = YGate()


class CYGate(BasicGate):
    """Control Y gate.

    Example:
        .. code-block:: python

            CY * (qr[0], qr[1])
    """

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "CY"

    def __or__(self, qubits):
        """Quantum logic gate operation.

        Args:
            qubits: qubits[0] is control qubit, qubits[1] is target qubit.

        Example:
            .. code-block:: python

                CY * (qr[0], qr[1])
        """
        if len(qubits) != 2:
            raise ValueError("Parameter error: One controlled and one target qubit are required.")

        self.qubits = qubits
        controls = [qubits[0].index]
        targets = [qubits[1].index]
        cmd = Command(self, targets, controls, inverse=self.is_inverse)
        self.commit(qubits[0].circuit, cmd)

    def __mul__(self, qubits):
        """Overwrite * operator to achieve quantum logic gate operation, reuse __or__ operator implement."""
        self.__or__(qubits)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.array([[0, 0, -1j, 0], [0, 1, 0, 0], [1j, 0, 0, 0], [0, 0, 0, 1]])

    def inv(self):
        """Apply inverse gate"""
        gate = CYGate()
        gate.is_inverse = not self.is_inverse
        return gate


CY = CYGate()



from qutrunk.circuit.gates import (
    MCX,
    CYGate,
    MCZ,
    SwapGate,
    HGate,
    SGate,
    SdgGate,
    SqrtXGate,
    TGate,
    TdgGate,
    P,
    U1,
    U2,
    U3,
    U,
    CU,
    CU1,
    CU3,
    XGate,
    YGate,
    ZGate,
    Rx,
    Ry,
    Rz,
    Rxx,
    Rzz,
    CP,
    CRx,
    CRy,
    CRz,
    R,
    Ryy,
    SqrtSwapGate,
    iSwapGate,
    CR,
    X1Gate,
    Y1Gate,
    Z1Gate,
    IGate,
    SqrtXdgGate,
    CHGate,
    CSqrtXGate,
    CSwapGate,
)

# qasm的量子门映射到qutrunk的量子门
qutrunk_standard_gate = {
    # standard gate from qasm
    "u1": U1,
    "u2": U2,
    "u3": U3,
    "u": U,
    "p": P,
    "x": XGate,
    "y": YGate,
    "z": ZGate,
    "t": TGate,
    "tdg": TdgGate,
    "s": SGate,
    "sdg": SdgGate,
    "sx": SqrtXGate,
    "sxdg": SqrtXdgGate,
    "swap": SwapGate,
    "rx": Rx,
    "rxx": Rxx,
    "ry": Ry,
    "rz": Rz,
    "rzz": Rzz,
    "id": IGate,
    "h": HGate,
    "cx": MCX,
    "cy": CYGate,
    "cz": MCZ,
    "ch": CHGate,
    "crx": CRx,
    "cry": CRy,
    "crz": CRz,
    "csx": CSqrtXGate,
    "cu1": CU1,
    "cp": CP,
    "cu": CU,
    "cu3": CU3,
    # "ccx": MCX(2), # qutrunk implement by MCX
    "cswap": CSwapGate,
    # "delay": Delay,
    # qutrunk only gate
    "r": R,
    "ryy": Ryy,
    "sqrtswap": SqrtSwapGate,
    "iswap": iSwapGate,
    "cr": CR,
    "x1": X1Gate,
    "y1": Y1Gate,
    "z1": Z1Gate,
    "mcx": MCX,
    "mcz": MCZ,
}



"""Self-increment operation."""

from qutrunk.circuit import Qureg
from qutrunk.circuit.ops.operator import Operator
from qutrunk.circuit.gates import X, MCX


class Increment(Operator):
    """Self-increment operation.

    Example:
        .. code-block:: python

            from qutrunk.circuit import QCircuit
            from qutrunk.circuit.gates import Measure, All
            from qutrunk.circuit.ops import Classical
            from qutrunk.circuit.ops import INC

            circuit = QCircuit()
            qr = circuit.allocate(4)
            Classical(0) * qr
            INC * qr
            All(Measure) * qr
            res = circuit.run()
            print(res.get_outcome())

    """

    def __init__(self):
        super().__init__()

    def __mul__(self, qr: Qureg):
        if not isinstance(qr, Qureg):
            raise TypeError("The operand must be Qureg.")

        num_qubits = len(qr)

        ctrl = []
        for i in range(num_qubits, 1, -1):
            for j in range(i - 1):
                ctrl.append(qr[j])
            MCX(i - 1) * (*ctrl, qr[i - 1])
            ctrl = []

        X * qr[0]


INC = Increment()



from qutrunk.converters.circuit_to_dag import circuit_to_dag


def _get_instructions(circuit):
    """Given a circuit, return a tuple(qubits, cbits, nodes).

    Args:
        circuit: A quantum circuit.

    Returns:
        qubits: A list of QuBit.
        cbits: A list of CBit.
        nodes: Nodes is a list of DAG nodes whose type is "operation".
            (qubits, cbits, nodes)
    """

    dag = circuit_to_dag(circuit)

    qubits = dag.qubits
    cbits = dag.cbits

    nodes = []
    for node in dag.topological_op_nodes():
        nodes.append([node])
    nodes = [
        [node for node in layer if any(q in qubits for q in node.qargs)]
        for layer in nodes
    ]

    return qubits, cbits, nodes



"""Parameter Class for variable parameters."""

from uuid import uuid4


class Parameter:
    """Parameter Class for variable parameters.

    Args:
        name(str): the name of Parameter.
    """

    def __new__(cls, name):
        obj = object.__new__(cls)
        # TODO:??
        obj._uuid = uuid4()
        obj._hash = hash(obj._uuid)

        return obj

    def __init__(self, name):
        self.name = name
        self.value = None
        self._host = None

    def __copy__(self):
        return self

    def __deepcopy__(self, memo=None):
        return self

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if isinstance(other, Parameter):
            return self._uuid == other._uuid
        else:
            return False

    def __getstate__(self):
        return {"name": self.name, "value": self.value}

    def __setstate__(self, state):
        self.name = state["name"]
        self.value = state["value"]

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"

    @property
    def host(self):
        """Get parameter host."""
        return self._host

    @host.setter
    def host(self, host):
        """Set parameter host.
        
        Args:
            host: The Host using this parameter. 
        """
        self._host = host

    def update(self, value):
        """Update parameter value.
        
        Args:
            value: Parameter value.
        """
        self.value = value
        if self._host:
            self._host.update_parameter(self)




import numpy as np

from qutrunk.circuit import Command
from .basicgate import BasicGate, Observable, PauliType


class IGate(BasicGate, Observable):
    """Apply the single-qubit Identity gate.

    Example:
        .. code-block:: python

            I * qr[0]
    """

    def __init__(self):
        super().__init__()

    def __str__(self):
        return "Id"

    def __or__(self, qubit):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to apply I gate.

        Example:
            .. code-block:: python

                I * qr[0]
        """
        targets = [qubit.index]
        cmd = Command(self, targets, inverse=self.is_inverse)
        self.commit(qubit.circuit, cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, \
            reuse __or__ operator implement."""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        return np.matrix([[1, 0], [0, 1]])

    def __call__(self, target):
        """
        Get Observable data.

        Args:
            target: The observed qubit.

        Returns:
            The observed data list, each item contains op type and target qubit, \
                e.g: [{"oper_type": 1, "target": 0}].
        """
        pauli = {}
        pauli["oper_type"] = PauliType.PAULI_I.value
        pauli["target"] = target.index
        return pauli

    def inv(self):
        """Apply inverse gate."""
        gate = IGate()
        gate.is_inverse = not self.is_inverse 
        return gate


PauliI = I = IGate()



"""
Contains ops for qutrunk.
"""

from .operator import Operator
from .qaa import QAA
from .qft import IQFT, QFT
from .amp import AMP
from .plus import PLUS
from .classical import Classical
from .increment import INC
from .decrement import DEC
from .addition import ADD
from .qpe import QPE



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Node for an OPENQASM custom gate statement."""
from .node import Node


class CustomUnitary(Node):
    """Node for an OPENQASM custom gate statement.

    children[0] is an id node.
    children[1] is an exp_list (if len==3) or primary_list.
    children[2], if present, is a primary_list.

    Has properties:
    .id = id node
    .name = gate name string
    .arguments = None or exp_list node
    .bitlist = primary_list node
    """

    def __init__(self, children):
        """Create the custom gate node."""
        super().__init__("custom_unitary", children, None)
        self.id = children[0]  # pylint: disable=invalid-name
        self.name = self.id.name
        if len(children) == 3:
            self.arguments = children[1]
            self.bitlist = children[2]
        else:
            self.arguments = None
            self.bitlist = children[1]

    def qasm(self):
        """Return the corresponding OPENQASM string."""
        string = self.name
        if self.arguments is not None:
            string += "(" + self.arguments.qasm() + ")"
        string += " " + self.bitlist.qasm() + ";"
        return string



"""Command Module."""

from qutrunk.circuit.parameter import Parameter


class Amplitude:
    """Set state-vector Amplitude.

    Args:
        reals: Amplitude read part.
        imags: Amplitude imag part.
        startind: Amplitude start index.
        numamps: Amplitude number.
    """

    def __init__(self):
        self.reals = []
        self.imags = []
        self.startind = 0
        self.numamps = 0


class Mat:
    """Custom matrix.

    Args:
        reals: matrix read part.
        imags: matrix imag part.
        unitary: matrix is unitary or not
    """
    def __init__(self):
        self.reals = []
        self.imags = []
        self.unitary = False


class CmdEx:
    """Command extension.

    Args:
        amp: Amplitude object.
    """
    def __init__(self, amp=None, mat=None):
        self.amp = amp
        self.mat = mat


class Command:
    """Converts the quantum gate operation into a specific command.

    Args:
        gate: Quantum gate operator.
        targets: Target qubits.
        controls: Control qubits.
        rotation: Angle for rotation gate.
        inverse: Whether to enable the inverse circuit.
    """

    def __init__(
        self,
        gate,
        targets=None,
        controls=None,
        rotation=None,
        inverse=False,
        cmdex=None,
    ):
        # TODO: modify controls and rotation to tuple?
        if targets is None:
            self.targets = []
        else:
            self.targets = list(targets)

        if controls is None:
            self.controls = []
        else:
            self.controls = list(controls)

        if rotation is None:
            self.rotation = []
        else:
            self.rotation = list(rotation)

        self.gate = gate
        self.cmd_ver = "1.0"
        self.inverse = inverse

        self.parameters = {}
        for i, r in enumerate(self.rotation):
            if isinstance(r, Parameter):
                # map index to parameter
                self.parameters[i] = r
                r.host = self

        # Command extention data
        # TODO: extra
        self.cmdex = cmdex

    def __eq__(self, other):
        """Two command are the same if they have the same qasm."""
        # TODO: need to improve
        if type(self) is not type(other):
            return False

        if self.qasm() == other.qasm():
            return True
        return False

    def __repr__(self) -> str:
        """Generate a representation of the command object instance.

        Returns:
            str: A representation of the command instance.
        """
        return (
            f"Command(gate={self.gate}, controls={self.controls}, targets={self.targets}, "
            f"rotation={self.rotation}), inverse={self.inverse})"
        )

    def qasm(self) -> str:
        """Generate OpenQASM code for command."""
        name = str(self.gate).lower()

        if str(self.gate) == "Measure":
            index = self.targets[0]
            return f"{name} q[{str(index)}] -> c[{str(index)}]"

        ctrl_cnt = len(self.controls)

        # OpenQASM use cx/ccx/c3x..., qutrunk use mcx/mcx(n)
        if name.startswith("mc"):
            name = name.replace("mc", "c", 1)

        if ctrl_cnt == 2:
            name = name.replace("c", "cc", 1)
        elif ctrl_cnt > 2:
            name = name.replace("c", "c" + str(ctrl_cnt), 1)

        angles_str = ""
        if len(self.rotation) > 0:
            angles_str = "(" + ",".join([str(ag) for ag in self.rotation]) + ")"

        qubits_index = self.controls + self.targets
        qubits_str = ",".join([f"q[{qi}]" for qi in qubits_index])
        return name + angles_str + " " + qubits_str

    def qusl(self) -> str:
        """Generate QuSL code for command."""
        name = str(self.gate)
        if name == "AMP":
            return f"AMP({self.gate.classicvector}, {self.gate.startind}, {self.gate.numamps}) * q"
            # return 'AMP({}, {}, {}) * q'.format(self.gate.classicvector, self.gate.startind, self.gate.numamps)

        params = []
        param_str = ""
        inv_str = ""

        # only append control bit count as param when it's more than one
        if self.cmdex and self.cmdex.mat:
            params += self.gate.matrix

        ctrl_cnt = len(self.controls)
        if ctrl_cnt:
            params.append(ctrl_cnt)

        if len(self.rotation) > 0:
            params += self.rotation

        if params:
            param_str = "(" + ", ".join([str(param) for param in params]) + ")"

        qubits_index = self.controls + self.targets
        qubits_str = ", ".join([f"q[{qi}]" for qi in qubits_index])

        # add parentheses when qubits count is more than one
        if len(qubits_index) > 1:
            qubits_str = "(" + qubits_str + ")"

        if self.inverse:
            inv_str += ".inv()"

        return name + param_str + inv_str + " * " + qubits_str

    @property
    def name(self) -> str:
        """Command name."""
        return self.gate

    @property
    def num_qubits(self) -> int:
        """Return the number of qubits."""
        return len(self.controls) + len(self.targets)

    def update_parameter(self, param):
        """Update command parameter.

        Usually, parameter is the angle of rotation gate

        Args:
            param: Parameter object holds parameter's name and value.
        """
        for k, v in self.parameters.items():
            if v == param:
                self.rotation[k] = v.value



"""U2 gate."""
import numpy as np

from .basicgate import BasicGate
from qutrunk.circuit import Command
from qutrunk.circuit.qubit import QuBit


class U2(BasicGate):
    """U2 gate.

    Args:
        theta: U2 gate parameter1.
        phi: U2 gate parameter2.

    Example:
        .. code-block:: python

            U2(pi/2, pi/2) * qr[0]
    """

    def __init__(self, theta, phi):
        """
        Args:
            theta: U2 gate parameter1.
            phi: U2 gate parameter2.
        """
        if theta is None or phi is None:
            raise ValueError("The argument cannot be empty.")
        super().__init__()
        self.theta = theta
        self.phi = phi

    def __str__(self):
        return "U2"

    def __or__(self, qubit):
        """Quantum logic gate operation.

        Args:
            qubit: The quantum bit to apply U2 gate.

        Example:
            .. code-block:: python

                U2(pi/2, pi/2) * qr[0]

        Raises:
            TypeError: If the argument is not a Qubit object.
        """
        if not isinstance(qubit, QuBit):
            raise TypeError("The argument must be Qubit object.")

        targets = [qubit.index]
        cmd = Command(
            self, targets, rotation=[self.theta, self.phi], inverse=self.is_inverse
        )
        self.commit(qubit.circuit, cmd)

    def __mul__(self, qubit):
        """Overwrite * operator to achieve quantum logic gate operation, \
            reuse __or__ operator implement."""
        self.__or__(qubit)

    @property
    def matrix(self):
        """Access to the matrix property of this gate."""
        isqrt2 = 1 / np.sqrt(2)
        phi = self.theta
        lam = self.phi
        return np.matrix(
            [
                [isqrt2, -np.exp(1j * lam) * isqrt2],
                [np.exp(1j * phi) * isqrt2, np.exp(1j * (phi + lam)) * isqrt2],
            ]
        )

    def inv(self):
        """Apply inverse gate."""
        gate = U2(self.theta, self.phi)
        gate.is_inverse = not self.is_inverse
        return gate
